/* <LIC_AMD_STD>
 * Copyright (c) 2004-2005 Advanced Micro Devices, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * The full GNU General Public License is included in this distribution in the
 * file called COPYING
 * </LIC_AMD_STD>  */
/* <CTL_AMD_STD>
 * </CTL_AMD_STD>  */
/* <DOC_AMD_STD>
 *  CS5535 documentation available from AMD.
 * </DOC_AMD_STD>  */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ide.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/msr.h>

#include "ide-timing.h"
#include "ide_modes.h"
#include "cs5535.h"

#if defined(DISPLAY_CS5535_TIMINGS) && defined(CONFIG_PROC_FS)
#include <linux/stat.h>
#include <linux/proc_fs.h>

static u8 cs5535_proc = 0;
static struct pci_dev *bmide_dev;
static u8 w80 = 0;

static int cs5535_get_info (char *buffer, char **addr, off_t offset, int count)
{
	char *p = buffer;
	unsigned long bibma = pci_resource_start(bmide_dev, 4);

	u8  c0 = 0;

	/*
	 * at that point bibma+0x2 et bibma+0xa are byte registers
	 * to investigate:
	 */

	c0 = inb_p((u16)bibma + 0x02);

	p += sprintf(p, "\n                                "
		     "AMD/NS 5535 Chipset.\n");

	p += sprintf(p, "--------------- drive0 --------- drive1 ------\n");
	p += sprintf(p, "DMA enabled:    %s              %s\n",(c0&0x20) ? "yes" : "no ",
		     (c0&0x40) ? "yes" : "no ");

	pci_read_config_byte(bmide_dev, CS5535_CABLE_DETECT, &c0);

	p += sprintf(p, "80 Wire Cable:  %s              %s\n", (c0 & 1) ? "yes" : "no ", 
		     (c0 & 2) ? "yes" :  "no ");
	
	return p-buffer;
}

#endif /* DISPLAY_CS5535_TIMINGS && CONFIG_PROC_FS */
  
/* Format I PIO settings. We seperate out cmd and data for safer timings */

static unsigned int cs5535_pio_cmd_timings[5] = 
{ 0xF7F4, 0x53F3, 0x13F1, 0x5131, 0x1131 };
static unsigned int cs5535_pio_dta_timings[5] = 
{ 0xF7F4, 0xF173, 0x8141, 0x5131, 0x1131 };

static unsigned int cs5535_mwdma_timings[3] = 
{ 0x7F0FFFF3, 0x7F035352, 0x7f024241 };

static unsigned int cs5535_udma_timings[5] = 
{ 0x7F7436A1, 0x7F733481, 0x7F723261, 0x7F713161, 0x7F703061 };

/* Macros to check if the register is the reset value -  reset value is an
   invalid timing and indicates the register has not been set previously */

#define CS5535_BAD_PIO(timings) ( (timings&~0x80000000)==0x00009172 )
#define CS5535_BAD_DMA(timings) ( (timings & 0x000FFFFF) == 0x00077771 )

#define DEFAULT_PIO   ( (cs5535_pio_cmd_timings[0] << 16) | cs5535_pio_dta_timings[0] )

static u8 cs5535_get_dma_speed(u32 timings) {
	int i;

	for(i = 0; i < 3; i++) 
		if (cs5535_mwdma_timings[i] == timings)
			return  XFER_MW_DMA_0 + i;
	
	for(i = 0; i < 5; i++)
		if (cs5535_udma_timings[i] == timings)
			return XFER_UDMA_0 + i;

	return -1;
}
	
static u8 cs5535_get_pio_mode(int unit) {

	u32 reg, dummy;
	int i;

	rdmsr(unit ? ATAC_CH0D1_PIO : ATAC_CH0D0_PIO, reg, dummy);
	
	for(i = 0; i < 5; i++) {
		if ((reg & 0xFFFF) == cs5535_pio_dta_timings[i])
			return i;
	}

	BUG();
	return 0;
}

static void cs5535_set_speed(ide_drive_t *drive, u8 speed)  {

	u32 reg, dummy;
	int unit = drive->select.b.unit;

	if ((speed & XFER_MODE) == XFER_PIO) {
		u8 data1 = speed - XFER_PIO_0;
		u8 data2 = cs5535_get_pio_mode(!unit);

		u8 cmd = (data1 < data2) ? data1 : data2;

		/* Write the other drive timings */
		reg = (cs5535_pio_cmd_timings[cmd] << 16) | cs5535_pio_dta_timings[data2];
		wrmsr(!unit ? ATAC_CH0D1_PIO : ATAC_CH0D0_PIO, reg, 0);

		/* Write our drive timings*/
		reg = (cs5535_pio_cmd_timings[cmd] << 16) | cs5535_pio_dta_timings[data1];
		wrmsr(unit ? ATAC_CH0D1_PIO : ATAC_CH0D0_PIO, reg, 0);

		/* Write the Format 1 bit in the DMA register */
		rdmsr(unit ? ATAC_CH0D1_DMA : ATAC_CH0D0_DMA, reg, dummy);    
		wrmsr(unit ? ATAC_CH0D1_DMA : ATAC_CH0D0_DMA, reg | 0x80000000, 0);
	}
	else {
		rdmsr(unit ? ATAC_CH0D1_DMA : ATAC_CH0D0_DMA, reg, dummy);
		
		
		if (speed >= XFER_UDMA_0 && speed <= XFER_UDMA_7) 
			reg = (reg & 0x80000000) | cs5535_udma_timings[speed - XFER_UDMA_0];
		else if (speed >= XFER_MW_DMA_0 && speed <= XFER_MW_DMA_2) 
			reg = (reg & 0x80000000) | cs5535_mwdma_timings[speed - XFER_MW_DMA_0];
		else 
			return;

		wrmsr(unit ? ATAC_CH0D1_DMA : ATAC_CH0D0_DMA, reg, 0);
	}
}

static int cs5535_set_drive(ide_drive_t *drive, u8 speed) 
{       
	if (speed != drive->current_speed)
		ide_config_drive_speed(drive, speed);
	
	cs5535_set_speed(drive, speed);	
	drive->current_speed = speed;
	return 0;
}


static void cs5535_tuneproc (ide_drive_t *drive,u8 pio) 
{ 	
	u8 speed = 0;

 	if (pio == 255)
		speed = ide_find_best_mode(drive, XFER_PIO | XFER_EPIO);
	else if (pio < 100) 
		speed = XFER_PIO_0 + min_t(byte, pio, 5);
	else if (pio < 200)
		speed = XFER_MW_DMA_0 + min_t(byte, (pio - 100), 2);
	else 
		speed = XFER_UDMA_0 + min_t(byte, (pio - 200), 7);
			
	cs5535_set_drive(drive, speed);
}

static int cs5535_config_dma(ide_drive_t *drive) {

	u8 speed = 0;
	unsigned long map = XFER_MWDMA | XFER_UDMA;
	if (w80) map |= XFER_UDMA_66;

	/* Get the best mode for this drive */
	speed = ide_find_best_mode(drive, map);
		
	if ((speed & XFER_MODE) != XFER_PIO) {
		cs5535_set_drive(drive, speed);
		
		if (drive->autodma)
			return HWIF(drive)->ide_dma_on(drive);
	}
	
	return 
		HWIF(drive)->ide_dma_off_quietly(drive);
}
	
static unsigned int __init init_chipset_cs5535 (struct pci_dev *dev, const char *name)
{
	unsigned long flags;
	u8 bit;

#if defined(DISPLAY_CS5535_TIMINGS) && defined(CONFIG_PROC_FS)
	if (!cs5535_proc)
	{
		cs5535_proc = 1;
		bmide_dev = dev;
		ide_pci_register_host_proc(&cs5535_procs[0]);
	}
#endif /* DISPLAY_CS5535_TIMINGS && CONFIG_PROC_FS */

	pci_read_config_byte(bmide_dev, CS5535_CABLE_DETECT, &bit);
	w80 = bit & 1;  /* Remember if we have a 80 wire cable */

	return 0;
}

static void __init init_hwif_cs5535 (ide_hwif_t *hwif)
{
unsigned long dummy;
u32           timings;
unsigned long dfltpio;
unsigned long bibma;
int           dmastatus;

	hwif->autodma = 0;
	dfltpio = (cs5535_pio_cmd_timings[0] << 16) | 
	cs5535_pio_dta_timings[0];
	dmastatus = 0;

	if (!bmide_dev)
		return;
	else
    	{
		bibma = pci_resource_start(bmide_dev, 4);
		dmastatus = inb_p((u16) bibma + 0x02);		
	}

	if (hwif->mate)
		hwif->serialized = hwif->mate->serialized = 1;

	hwif->tuneproc = &cs5535_tuneproc;
	hwif->speedproc = &cs5535_set_drive;

	/* Check the registers, and set up the PIO registers if nessesary */
	/* Note that this will default to whatever the BIOS set them to be */

	/* Drive 0 */

 	rdmsr(ATAC_CH0D0_PIO,timings,dummy);
	if (CS5535_BAD_PIO(timings)) 
	{
	 	/* PIO timings not initialized? */
		wrmsr(ATAC_CH0D0_PIO, dfltpio, 0);

		if(!hwif->drives[0].autotune)
			hwif->drives[0].autotune = 1;
	}

	/* Drive 1 */

	rdmsr(ATAC_CH0D1_PIO,timings,dummy);
	if (CS5535_BAD_PIO(timings)) 
	{
		/* PIO timings not initialized? */
		wrmsr(ATAC_CH0D1_PIO,dfltpio,0);
		if(!hwif->drives[1].autotune)
			hwif->drives[1].autotune = 1;

	}

	/* We want to take special care not to mess with the settings
	   from the BIOS.  If the DMA registers are already set, just
	   assume those are correct, and go from there */

 	rdmsr(ATAC_CH0D0_DMA, timings, dummy);

	if (!CS5535_BAD_DMA(timings)) {
		if (dmastatus & 0x20)
			hwif->ide_dma_on(&(hwif->drives[0]));

		hwif->drives[0].autotune = 2; 
	}
	
	rdmsr(ATAC_CH0D1_DMA, timings, dummy);

	if (!CS5535_BAD_DMA(timings)) {
		if (dmastatus & 0x40)
			hwif->ide_dma_on(&(hwif->drives[1]));

		hwif->drives[1].autotune = 2;
	}

	hwif->atapi_dma = 1;
    
	hwif->ultra_mask = 0x1F;
	hwif->mwdma_mask = 0x07;

	/* w80 will be 1 if a 80 wire cable was detected */

	hwif->udma_four = w80;

	hwif->ide_dma_check = &cs5535_config_dma;
	if (!noautodma) hwif->autodma = 1;
		
	hwif->drives[0].autodma = hwif->autodma;
	hwif->drives[1].autodma = hwif->autodma;
}

extern void ide_setup_pci_device(struct pci_dev *, ide_pci_device_t *);
static int __devinit cs5535_init_one(struct pci_dev *dev, const struct pci_device_id *id)
{
	ide_pci_device_t *d = &cs5535_chipsets[id->driver_data];

	if (dev->device != d->device) BUG();

	ide_setup_pci_device(dev, d);
	MOD_INC_USE_COUNT;
	return 0;
}

/* NOTE:  This is not for customer release - this provides support for
   the Hiddensee A0 IDE device. 
*/

#ifndef PCI_DEVICE_ID_AMD_CS5536_A0IDE  
#define PCI_DEVICE_ID_AMD_CS5536_A0IDE    0x2092
#define PCI_DEVICE_ID_AMD_CS5536_SOIDE    0x209A
#endif

static struct pci_device_id cs5535_pci_tbl[] __devinitdata = 
{
	{ PCI_VENDOR_ID_NS, PCI_DEVICE_ID_NS_CS5535_IDE, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{ PCI_VENDOR_ID_AMD, PCI_DEVICE_ID_AMD_CS5536_SOIDE, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 1},
	{ 0, },
};

static struct pci_driver driver = {
	.name       = "CS5535/6 IDE",
	.id_table   = cs5535_pci_tbl,
	.probe      = cs5535_init_one,
};


static int cs5535_ide_init(void)
{
	return ide_pci_register_driver(&driver);
}

static void cs5535_ide_exit(void)
{
	ide_pci_unregister_driver(&driver);
}

module_init(cs5535_ide_init);
module_exit(cs5535_ide_exit);

MODULE_AUTHOR("AMD");
MODULE_DESCRIPTION("PCI driver module for AMD CS5535/6 IDE");
MODULE_LICENSE("GPL");

EXPORT_NO_SYMBOLS;
